import 'package:riverpod/riverpod.dart';
import 'package:endlich/endlich.dart';
import 'provider.dart';

import 'builder.dart';
import 'value_provider.dart';

abstract class AutoDisposeServiceProviderRef<Value, Event>
    implements AutoDisposeProviderRefBase {}

class AutoDisposeServiceProviderElement<Value, Event>
    extends AutoDisposeProviderElementBase<Service<Value, Event>>
    implements AutoDisposeServiceProviderRef<Value, Event> {
  AutoDisposeServiceProviderElement(
      AutoDisposeProviderBase<Service<Value, Event>> provider)
      : super(provider);
}

class AutoDisposeServiceProvider<Value, Event>
    extends AutoDisposeProviderBase<Service<Value, Event>> {
  AutoDisposeServiceProvider(this._create, {String? name})
      : super(name == null ? null : '$name.service');

  final Create<Service<Value, Event>,
      AutoDisposeServiceProviderRef<Value, Event>> _create;

  @override
  Service<Value, Event> create(
      AutoDisposeServiceProviderRef<Value, Event> ref) {
    final service = _create(ref);
    ref.onDispose(service.dispose);
    return service;
  }

  @override
  AutoDisposeServiceProviderElement<Value, Event> createElement() {
    return AutoDisposeServiceProviderElement<Value, Event>(this);
  }

  @override
  void setupOverride(SetupOverride setup) {
    throw UnsupportedError('Cannot override MachineStateProvider.service');
  }

  @override
  bool updateShouldNotify(
      Service<Value, Event> previousState, Service<Value, Event> newState) {
    return true;
  }
}

abstract class AutoDisposeMachineStateProviderRef<Value, Event>
    implements AutoDisposeProviderRefBase {}

class AutoDisposeMachineStateProviderElement<Value, Event>
    extends AutoDisposeProviderElementBase<MachineState<Value, Event>>
    implements AutoDisposeMachineStateProviderRef<Value, Event> {
  AutoDisposeMachineStateProviderElement(
      AutoDisposeProviderBase<MachineState<Value, Event>> provider)
      : super(provider);
}

class AutoDisposeMachineStateProvider<Value, Event>
    extends AutoDisposeProviderBase<MachineState<Value, Event>>
    with MachineStateProviderMixin<Value, Event> {
  AutoDisposeMachineStateProvider(this._create, {String? name}) : super(name);

  static const family = AutoDisposeMachineStateProviderFamilyBuilder();

  final Create<Service<Value, Event>,
      AutoDisposeServiceProviderRef<Value, Event>> _create;

  @override
  late final AutoDisposeServiceProvider<Value, Event> service =
      AutoDisposeServiceProvider<Value, Event>(_create, name: name);

  @override
  MachineState<Value, Event> create(
      AutoDisposeMachineStateProviderRef<Value, Event> ref) {
    final ele = ref as AutoDisposeMachineStateProviderElement<Value, Event>;
    final s = ref.watch(service);
    ref.onDispose(s.watch(
      (value) => ele.state = value,
      fireImmediately: true,
    ));
    return ele.state;
  }

  @override
  AutoDisposeProviderElementBase<MachineState<Value, Event>> createElement() {
    return AutoDisposeMachineStateProviderElement<Value, Event>(this);
  }

  @override
  bool updateShouldNotify(MachineState<Value, Event> previousState,
      MachineState<Value, Event> newState) {
    return true;
  }

  Override overrideWithProvider(
    AutoDisposeMachineStateProvider<Value, Event> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: service, override: provider.service);
      setup(origin: this, override: this);
    });
  }
}

class AutoDisposeMachineStateProviderFamily<Value, Event, Arg> extends Family<
    MachineState<Value, Event>,
    Arg,
    AutoDisposeMachineStateProvider<Value, Event>> {
  AutoDisposeMachineStateProviderFamily(this._create, {String? name})
      : super(name);

  final FamilyCreate<Service<Value, Event>,
      AutoDisposeServiceProviderRef<Value, Event>, Arg> _create;

  @override
  AutoDisposeMachineStateProvider<Value, Event> create(
    Arg argument,
  ) {
    final provider = AutoDisposeMachineStateProvider<Value, Event>(
      (ref) => _create(ref, argument),
      name: name,
    );

    registerProvider(provider.service, argument);

    return provider;
  }

  /// Overrides the behavior of a family for a part of the application.
  ///
  /// {@macro riverpod.overideWith}
  Override overrideWithProvider(
    AutoDisposeMachineStateProvider<Value, Event> Function(Arg argument)
        override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.service, override: override(arg).service);
        setup(origin: provider, override: provider);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.service, override: provider.service);
  }
}
