import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/src/provider.dart';
import 'package:test/test.dart';

import 'stuff.dart';

void main() {
  group('StateNotifier.family', () {
    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final service = TestService();
        final provider =
            MachineStateProvider.autoDisposeFamily<TestValue, TestEvent, int>(
          (ref, _) => service,
        );
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider(0).service), service);
        expect(container.read(provider(0)), isAValueOf(TestValue.foo()));
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).service),
          ]),
        );
        expect(root.getAllProviderElementsInOrder(), isEmpty);
      });

      test('when using provider.overrideWithProvider', () async {
        final controller = TestService();
        final provider =
            MachineStateProvider.autoDisposeFamily<TestValue, TestEvent, int>(
                (ref, _) => controller);
        final root = createContainer();
        final controllerOverride = TestService(initial: TestValue.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            (value) =>
                MachineStateProvider.autoDispose((ref) => controllerOverride),
          ),
        ]);

        expect(container.read(provider(0).service), controllerOverride);
        expect(container.read(provider(0)), isAValueOf(TestValue.bar()));
        expect(root.getAllProviderElementsInOrder(), isEmpty);
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).service),
          ]),
        );
      });
    });

    test('properly overrides ==', () {
      final family =
          MachineStateProvider.autoDisposeFamily<TestValue, TestEvent, int>(
              (ref, _) => TestService());

      expect(family(0), family(0));
      expect(family(1), isNot(family(0)));
      expect(family(1), family(1));
    });
  });
}
