import 'package:endlich/endlich.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/mockito.dart';
import 'package:riverpod/riverpod.dart';
import 'package:test/test.dart';

part 'stuff.freezed.dart';

@freezed
class TestValue with _$TestValue {
  factory TestValue.foo() = _Foo;
  factory TestValue.bar() = _Bar;
  factory TestValue.baz() = _Baz;
}

@freezed
class TestEvent with _$TestEvent {
  factory TestEvent.next() = _Next;
  factory TestEvent.toFoo() = _ToFoo;
  factory TestEvent.toBar() = _ToBar;
  factory TestEvent.toBaz() = _ToBaz;
}

class TestService extends Service<TestValue, TestEvent> {
  TestService({
    this.initial,
    this.onFoo,
    this.onBar,
    this.onBaz,
  }) : super();

  final TestValue? initial;

  @override
  TestValue get initialValue => initial ?? TestValue.foo();

  StateConfig<TestValue, TestEvent> Function(
      _Foo obj, StateConfig<TestValue, TestEvent> cfg)? onFoo;
  StateConfig<TestValue, TestEvent> Function(
      _Bar obj, StateConfig<TestValue, TestEvent> cfg)? onBar;
  StateConfig<TestValue, TestEvent> Function(
      _Baz obj, StateConfig<TestValue, TestEvent> cfg)? onBaz;

  @override
  StateConfig<TestValue, TestEvent> onChange(TestValue value) => value.map(
        foo: (obj) =>
            onFoo?.call(obj, StateConfig<TestValue, TestEvent>()) ??
            StateConfig<TestValue, TestEvent>(),
        bar: (obj) =>
            onBar?.call(obj, StateConfig<TestValue, TestEvent>()) ??
            StateConfig<TestValue, TestEvent>(),
        baz: (obj) =>
            onBaz?.call(obj, StateConfig<TestValue, TestEvent>()) ??
            StateConfig<TestValue, TestEvent>(),
      );
}

class StateListener extends Mock {
  void call(MachineState<TestValue, TestEvent> value);
}

class ControllerListenerMock extends Mock {
  void call(Service<TestValue, TestEvent>? value);
}

typedef VerifyOnly = VerificationResult Function<T>(
  Mock mock,
  T matchingInvocations,
);

TypeMatcher<MachineState<TestValue, TestEvent>> isAValueOf(TestValue value) {
  return isA<MachineState<TestValue, TestEvent>>()
      .having((s) => s.value, 'state.value', value);
}

/// Syntax sugar for:
///
/// ```dart
/// verify(mock()).called(1);
/// verifyNoMoreInteractions(mock);
/// ```
VerifyOnly get verifyOnly {
  final verification = verify;

  return <T>(mock, invocation) {
    final result = verification(invocation);
    result.called(1);
    verifyNoMoreInteractions(mock);
    return result;
  };
}

ProviderContainer createContainer({
  ProviderContainer? parent,
  List<Override> overrides = const [],
  List<ProviderObserver>? observers,
}) {
  final container = ProviderContainer(
    parent: parent,
    overrides: overrides,
    observers: observers,
  );
  addTearDown(container.dispose);
  return container;
}
